from __future__ import annotations

import logging
import signal
import threading
import time
from concurrent.futures import Future
from datetime import datetime

is_killed_handling: bool = False
is_killed_value: bool = False
is_killed_lock = threading.RLock()


def is_killed_reset():
    global is_killed_lock
    with is_killed_lock:
        global is_killed_value
        is_killed_value = False


def is_killed() -> bool:
    """
    Check the process is has killed request or not.
    This function handling the signal.SIGINT and signal.SIGTERM signals, and injectable directly by is_killed_value.
    """
    global is_killed_lock
    with is_killed_lock:
        global is_killed_handling
        if not is_killed_handling:
            signal.signal(signal.SIGINT, is_killed_handling_func)
            signal.signal(signal.SIGTERM, is_killed_handling_func)
            is_killed_handling = True
        global is_killed_value
        return is_killed_value


def is_killed_handling_func(*args):
    global is_killed_lock
    with is_killed_lock:
        global is_killed_value
        logging.info("Handle kill signal")
        is_killed_value = True


def sleep_kc(seconds: float, sleep_step: float = 0.1) -> bool:
    """
    Sleep for seconds and return True. If is_killed() is True, then stop it immediately, and return False.
    :param seconds: number of seconds to sleep
    :param sleep_step: minimum sleep between kill signal checking (Default is 0.1s)
    """
    start = datetime.now().timestamp()
    while datetime.now().timestamp() - start < seconds:
        time.sleep(sleep_step)
        if is_killed():
            return False
    return True


def await_all_futures(futures: list[Future], sleep_step: float = 0.1) -> bool:
    """
    Wait for all futures to complete.
    If is_killed() is True, then return from this function with False, otherwise return True.
    :param futures: a list of concurrent.futures.Future objects
    :param sleep_step: time to wait between each check
    :return: True if all futures are done, False otherwise
    """
    while not is_killed():
        if not sleep_kc(1, sleep_step=sleep_step):
            return False
        has_not_done = False
        for f in futures:
            if not f.done():
                has_not_done = True
                break
        if has_not_done:
            continue
        else:
            break
    return not is_killed()



from __future__ import annotations

import copy
import io
import logging
import os
import re
import shutil
from builtins import float
from datetime import datetime, timezone
from typing import IO

from gwbackupy.storage.storage_interface import (
    StorageInterface,
    Path,
    Data,
    LinkFilter,
    LinkInterface,
    LinkList,
)


# Directory structure v2
# /[<path>]
#    - <id>.<propname>=<propvalue>.<propname>=<propvalue>.<ext>
#    - 12345678.m=12345678.metadata=.deleted=.json
#    - 12345678.m=12345678.object=message.eml.gz
#


class FileLink(LinkInterface):
    """
    Link object for a file.
    The properties are stored in the file name.
    To avoid too many files in one directory, it groups by the original creation date of the object, but this is irrelevant for listing.
    """

    filename_parser = re.compile(
        r"^(?P<id>[^.]+?)(?P<properties>(?:\.[a-z0-9]+=[^.]*)*)\.(?P<extension>.+)$"
    )

    def __init__(self):
        self.__id: str | None = None
        self.__path: str | None = None
        self.__properties: dict[str, any] = {}
        self.__extension: str | None = None

    def fill(self, values: dict[str, any], replace: bool = False) -> FileLink:
        if "object_id" in values:
            self.__id = values["object_id"]
        if replace:
            self.__properties = {}
        if "deleted" in values:
            self.__properties[LinkInterface.property_deleted] = True
        if "object" in values:
            self.__properties[LinkInterface.property_object] = True
        if "metadata" in values:
            self.__properties[LinkInterface.property_metadata] = True
        if "extension" in values:
            self.__extension = values["extension"]
        if "mutation" in values:
            self.__properties[LinkInterface.property_mutation] = values["mutation"]
        if "path" in values:
            self.__path = values["path"]
        for key in values:
            if key not in [
                "object_id",
                "deleted",
                "object",
                "metadata",
                "extension",
                "mutation",
                "path",
            ]:
                value = values[key]
                if isinstance(value, str):
                    if value == "":
                        value = True
                    self.__properties[key] = value
        return self

    def get_file_path(self) -> str:
        path = self.__path + "/" + self.__id
        keys = list(self.__properties.keys())
        keys.sort()
        for k in keys:
            if self.__properties[k] is None:
                continue
            path += f".{k}="
            if self.__properties[k] is not True:
                path += str(self.__properties[k])
        if self.__extension is not None:
            path += f".{self.__extension}"

        return path

    @staticmethod
    def parse_file_name(filename: str) -> dict[str, any] | None:
        m = FileLink.filename_parser.search(filename)
        if m is None:
            return None
        result: dict[str, any] = {
            "object_id": m.group("id"),
            "extension": m.group("extension"),
        }
        properties = m.group("properties")
        if properties is None or properties == "":
            return result
        for prop in properties.strip(".").split("."):
            p, v = prop.split("=", 1)
            if v == "":
                v = True
            result[p] = v
        return result

    def mutation(self) -> str:
        return self.get_property(LinkInterface.property_mutation)

    def id(self) -> str:
        return self.__id

    def get_properties(self) -> dict[str, any]:
        return self.__properties

    def get_property(self, name: str, default: any = None) -> any:
        if name in self.__properties:
            return self.__properties[name]
        return default

    def has_property(self, name: str) -> bool:
        if self.get_properties().get(name, None) is not None:
            return True
        return False

    def set_properties(self, sets: dict[str, any], replace: bool = False) -> FileLink:
        if replace:
            self.__properties = sets
            return self
        for k, v in sets.items():
            self.__properties[k] = v
        return self

    def is_deleted(self) -> bool:
        return self.has_property(LinkInterface.property_deleted) is True

    def is_metadata(self) -> bool:
        return self.has_property(LinkInterface.property_metadata) is True

    def is_object(self) -> bool:
        return self.has_property(LinkInterface.property_object) is True

    def __repr__(self) -> str:
        return f"{self.__class__}#id:{self.id()},props:{self.__properties},path:{self.__path}"

    def __eq__(self, other):
        if not isinstance(other, FileLink):
            return False
        if self.id() != other.id():
            return False
        if self.get_properties() != other.get_properties():
            return False
        if self.__path != other.__path:
            return False
        if self.__extension != other.__extension:
            return False
        return True


class FileStorage(StorageInterface):
    def __init__(self, root: str):
        self.root = root

    def new_link(
        self,
        object_id: str,
        extension: str,
        created_timestamp: int | float | None = None,
    ) -> FileLink:
        link = FileLink()
        path = self.root
        if created_timestamp is not None:
            sub_paths = (
                datetime.fromtimestamp(created_timestamp, tz=timezone.utc)
                .strftime("%Y-%m-%d")
                .split("-", 1)
            )
            path += f"/{sub_paths[0]}/{sub_paths[1]}"
        link.fill(
            {
                "path": path,
                "object_id": object_id,
                "extension": extension,
                "mutation": FileStorage.__gen_mutation(),
            }
        )
        return link

    @staticmethod
    def __gen_mutation():
        return str(int(datetime.utcnow().timestamp() * 1000))

    def get(self, link: FileLink) -> IO[bytes]:
        file_path = link.get_file_path()
        return open(file_path, "rb")

    def put(self, link: FileLink, data: Data) -> bool:
        file_path = link.get_file_path()
        logging.debug(f"Put object {file_path}")
        result = self.__write(file_path=file_path, data=data)
        if result:
            logging.debug(f"{file_path} put successfully")
            return True
        logging.error(f"File put fail ({file_path})")
        return False

    def remove(self, link: FileLink, as_new_mutation: bool = True) -> bool:
        if not as_new_mutation:
            try:
                if os.path.exists(link.get_file_path()):
                    os.remove(link.get_file_path())
                return True
            except BaseException as e:
                logging.exception(f"Delete fail {link.get_file_path()} with error: {e}")
                return False

        dst = copy.deepcopy(link).fill(
            {
                "deleted": True,
                "mutation": self.__gen_mutation(),
            }
        )
        try:
            with self.get(link) as f:
                if not f:
                    logging.error(f"{link.get_file_path()} not found or not readable")
                    return False
                self.put(dst, f)
        except BaseException as e:
            logging.exception(
                f"Copy as new mutation is failed {link.get_file_path()} -> {dst.get_file_path()}: {e}"
            )
            return False
        return True

    def find(self, f: LinkFilter | None = None) -> LinkList[LinkInterface]:
        abspath = self.root
        skip_path = len(abspath)
        result: LinkList[LinkInterface] = LinkList([])
        for _path, _, filenames in os.walk(abspath):
            for file in filenames:
                file_path = os.path.join(_path, file)
                relpath = _path[skip_path:]
                link = FileLink()
                if len(relpath) > 0:
                    link.path = relpath
                m = FileLink.parse_file_name(file)
                if m is None:
                    continue
                m["path"] = _path
                link.fill(m)

                if "extension" not in m:
                    logging.debug(f"Unknown file without extension: {file_path}")
                    continue
                if m["extension"] == "tmp":
                    logging.debug(f"Temporary file {file_path}, remove it")
                    try:
                        os.remove(file_path)
                    except BaseException as e:
                        logging.exception(
                            f"Temporary file remove fail {file_path}: {e}"
                        )
                    continue

                if f is None or f(link):
                    result.append(link)
        return result

    @staticmethod
    def __remove(file_path: str) -> bool:
        try:
            exists = os.path.exists(file_path)
            if not exists:
                return True
        except BaseException as e:
            logging.exception(f"File exists check fail {file_path}: {e}")
            return False

        try:
            os.remove(file_path)
        except BaseException as e:
            logging.exception(f"File remove fail {file_path}: {e}")
            return False
        return True

    @staticmethod
    def __write(file_path: str, data: Data) -> bool:
        path = os.path.dirname(file_path)
        try:
            path_exists = os.path.exists(path)
            if not path_exists:
                try:
                    os.makedirs(path, exist_ok=True)
                except BaseException as e:
                    logging.exception(f"Directory create fail {path}: {e}")
                    return False
        except BaseException as e:
            logging.exception(f"Directory exists check fail {path}: {e}")
            return False

        file_path_tmp = f"{file_path}.tmp"
        try:
            try:
                with open(file_path_tmp, "wb") as f:
                    if isinstance(data, bytes):
                        f.write(data)
                    elif isinstance(data, str):
                        f.write(bytes(data, "utf-8"))
                    elif isinstance(data, io.BufferedReader):
                        f.write(data.read())
                    elif callable(data):
                        data(f)
                    else:
                        raise RuntimeError(f"Not supported data type: {type(data)}")
            except BaseException as e:
                logging.exception(f"Temporary file writing fail {file_path_tmp}: {e}")
                return False
            try:
                shutil.move(file_path_tmp, file_path)
            except BaseException as e:
                logging.exception(
                    f"File rename fail {file_path_tmp} -> {file_path}: {e}"
                )
                return False
            return True
        finally:
            FileStorage.__remove(file_path_tmp)



from __future__ import annotations

import base64
import random
import string
from datetime import datetime
import json
import logging
from json import JSONDecodeError
from typing import IO

import tzlocal
from googleapiclient.errors import HttpError


def str_trim(text, length, postfix="..."):
    """
    Trim a string to a specified length, and add a postfix if the string is longer than the specified length.
    :param text: the string to be trimmed
    :param length: the maximum length of the string
    :param postfix: the postfix to add if the string is longer than the specified length (default is "...")
    :return: the trimmed string
    """
    if len(text) > length:
        text = text[:length] + postfix
    return text


def decode_base64url(data):
    padding = 4 - (len(data) % 4)
    data = data + ("=" * padding)
    return base64.urlsafe_b64decode(data)


def encode_base64url(data):
    return base64.urlsafe_b64encode(data).decode("utf-8").replace("=", "")


def parse_date(date: str, tz: tzlocal) -> datetime:
    """
    Parse a date string and return a datetime object with the corresponding timezone.
    :param date: string representing a date in the format "YYYY-MM-DD" or 'YYYY-MM-DD hh:mm:ss'
    :param tz: timezone to be set for the returned datetime object
    :return: datetime object with the corresponding timezone
    """
    tzs = datetime.now().astimezone(tz).strftime("%z")
    df = "%Y-%m-%d %H:%M:%S %z"
    try:
        return datetime.strptime(f"{date} {tzs}", df).astimezone(tz)
    except ValueError:
        try:
            return datetime.strptime(f"{date} 00:00:00 {tzs}", df).astimezone(tz)
        except ValueError:
            raise ValueError(f"Date time parsing failed: {date}")


def json_load(io: IO[bytes]) -> list[any] | dict[str, any] | None:
    try:
        return json.load(io)
    except JSONDecodeError as e:
        logging.exception(f"Invalid JSON format: {e}")
    return None


def is_rate_limit_exceeded(e) -> bool:
    if not isinstance(e, HttpError):
        return False
    e: HttpError
    if e.status_code != 403:
        return False
    d = e.error_details
    if not isinstance(d, list):
        return False
    if len(d) != 1:
        return False
    item = d[0]
    return (
        item.get("domain") == "usageLimits"
        and item.get("reason") == "rateLimitExceeded"
    )


def random_string(length: int = 8) -> str:
    return "".join(random.choice(string.ascii_lowercase) for _ in range(length))



import argparse
import logging
import sys
import threading

import pytz as pytz
from tzlocal import get_localzone

import gwbackupy.global_properties as global_properties
from gwbackupy.filters.gmail_filter import GmailFilter
from gwbackupy.gmail import Gmail
from gwbackupy.helpers import parse_date
from gwbackupy.providers.gapi_gmail_service_wrapper import GapiGmailServiceWrapper
from gwbackupy.providers.gmail_service_provider import GmailServiceProvider
from gwbackupy.storage.file_storage import FileStorage

lock = threading.Lock()


def parse_arguments() -> argparse.Namespace:
    log_levels = {
        "finest": global_properties.log_finest,
        "debug": logging.DEBUG,
        "info": logging.INFO,
        "warning": logging.WARN,
        "error": logging.ERROR,
        "critical": logging.CRITICAL,
    }

    parser = argparse.ArgumentParser(
        description="Google Workspace Backup Tool " + global_properties.version,
        add_help=False,
    )
    parser.set_defaults(feature=True)
    parser.add_argument(
        "-h",
        "--help",
        action="help",
        default=argparse.SUPPRESS,
        help="Show this help message and exit.",
    )
    parser.add_argument(
        "--dry", default=False, help="Run in dry mode", action="store_true"
    )
    parser.add_argument(
        "--timezone",
        type=lambda s: pytz.timezone(s),
        help="time zone",
        default=get_localzone(),
    )
    parser.add_argument(
        "--log-level",
        type=str.lower,
        help="Logging level: {keys}".format(keys=", ".join(log_levels.keys())),
        default="info",
        choices=log_levels,
    )
    parser.add_argument("--batch-size", type=int, help="Concurrent threads", default=5)
    parser.add_argument(
        "--service-account-email",
        type=str,
        default=None,
        help="Email of the service account (required for p12 keyfile)",
    )
    parser.add_argument(
        "--service-account-key-filepath",
        type=str,
        help="Path to the service account key file",
        required=False,
    )
    parser.add_argument(
        "--credentials-filepath",
        type=str,
        help="Path to the credentials file",
        required=False,
    )
    parser.add_argument(
        "--workdir",
        type=str,
        help="Path to the workdir",
        required=False,
        default="./data",
    )
    service_parser = parser.add_subparsers(dest="service")
    gmail_parser = service_parser.add_parser("gmail", help="GMail service commands")
    gmail_command_parser = gmail_parser.add_subparsers(dest="command")

    gmail_backup_parser = gmail_command_parser.add_parser("backup", help="Backup gmail")
    gmail_backup_parser.add_argument(
        "--email", type=str, help="Email of the account", required=True
    )
    gmail_backup_parser.add_argument(
        "--quick-sync-days",
        type=int,
        default=None,
        help="Quick sync number of days back. (It does not delete messages from local "
        "storage.)",
    )

    gmail_restore_parser = gmail_command_parser.add_parser(
        "restore", help="Restore gmail"
    )
    gmail_restore_parser.add_argument(
        "--email", type=str, help="Email from which restore", required=True
    )
    gmail_restore_parser.add_argument(
        "--to-email",
        type=str,
        help="Destination email account, if not specified, then --email is used",
    )
    gmail_restore_parser.add_argument(
        "--add-label",
        type=str,
        action="append",
        help="Add label to restored emails",
        default=None,
        dest="add_labels",
    )
    gmail_restore_parser.add_argument(
        "--restore-deleted",
        help="Restore deleted emails",
        default=False,
        action="store_true",
    )
    gmail_restore_parser.add_argument(
        "--restore-missing",
        help="Restore missing emails",
        default=False,
        action="store_true",
    )
    gmail_restore_parser.add_argument(
        "--filter-date-from",
        type=str,
        help="Filter date from (inclusive, format: yyyy-mm-dd or yyyy-mm-dd hh:mm:ss)",
        default=None,
    )
    gmail_restore_parser.add_argument(
        "--filter-date-to",
        type=str,
        help="Filter date to (exclusive, format: yyyy-mm-dd or yyyy-mm-dd hh:mm:ss)",
        default=None,
    )
    if len(sys.argv) == 1 or "--help" in sys.argv:
        parser.print_help(sys.stderr)
        sys.exit(1)
    args = parser.parse_args(args=sys.argv[1:])
    Log_Format = "%(levelname)s %(asctime)s - %(message)s"
    logging.addLevelName(global_properties.log_finest, "FINEST")
    logging.basicConfig(
        # filename="logfile.log",
        stream=sys.stdout,
        filemode="w",
        format=Log_Format,
        level=log_levels[args.log_level],
    )
    if (
        args.credentials_filepath is None and args.service_account_key_filepath is None
    ) or (
        args.credentials_filepath is not None
        and args.service_account_key_filepath is not None
    ):
        parser.error(
            "at least one of --credentials-filepath and --service-account-key-filepath required"
        )
    return args


def cli_startup():
    try:
        args = parse_arguments()
        if args.service == "gmail":
            storage = FileStorage(args.workdir + "/" + args.email + "/gmail")
            service_provider = GmailServiceProvider(
                credentials_file_path=args.credentials_filepath,
                service_account_email=args.service_account_email,
                service_account_file_path=args.service_account_key_filepath,
                storage=storage,
            )
            service_wrapper = GapiGmailServiceWrapper(
                service_provider=service_provider,
                dry_mode=args.dry,
            )
            gmail = Gmail(
                email=args.email,
                service_wrapper=service_wrapper,
                batch_size=args.batch_size,
                storage=storage,
                dry_mode=args.dry,
            )
            if args.command == "backup":
                if gmail.backup(quick_sync_days=args.quick_sync_days):
                    exit(0)
                else:
                    exit(1)
            if args.command == "restore":
                if args.add_labels is None:
                    args.add_labels = ["gwbackupy"]
                item_filter = GmailFilter()
                if args.restore_deleted:
                    item_filter.is_deleted()
                    logging.info("Filter options: deleted")
                if args.restore_missing:
                    item_filter.is_missing()
                    logging.info("Filter options: missing")
                if args.filter_date_from is not None:
                    dt = parse_date(args.filter_date_from, args.timezone)
                    item_filter.date_from(dt)
                    logging.info(f"Filter options: date from {dt}")
                if args.filter_date_to is not None:
                    dt = parse_date(args.filter_date_to, args.timezone)
                    item_filter.date_to(dt)
                    logging.info(f"Filter options: date to {dt}")
                if gmail.restore(
                    to_email=args.to_email,
                    item_filter=item_filter,
                    restore_deleted=args.restore_deleted,
                    restore_missing=args.restore_missing,
                    add_labels=args.add_labels,
                ):
                    exit(0)
                else:
                    exit(1)
            else:
                raise Exception("Unknown command")
    except KeyboardInterrupt:
        logging.warning("Process is interrupted")
        exit(1)
    except SystemExit as e:
        exit(e.code)
    except BaseException:
        logging.exception("CLI startup/run failed")
        exit(1)


if __name__ == "__main__":
    cli_startup()



from __future__ import annotations

from gwbackupy.providers.gapi_service_provider import GapiServiceProvider
from gwbackupy.storage.storage_interface import StorageInterface


class GmailServiceProvider(GapiServiceProvider):
    """Gmail service provider from gmail/v1 API with full access scope"""

    def __init__(
        self,
        credentials_file_path: str | None = None,
        service_account_file_path: str | None = None,
        service_account_email: str | None = None,
        storage: StorageInterface | None = None,
    ):
        super().__init__(
            "gmail",
            "v1",
            ["https://mail.google.com/"],
            credentials_file_path=credentials_file_path,
            storage=storage,
            service_account_file_path=service_account_file_path,
            service_account_email=service_account_email,
        )



from __future__ import annotations

from typing import Callable, IO, Union

Path = str
DataCall = Callable[[IO[bytes]], None]
Data = Union[str, bytes, DataCall, IO]


class LinkInterface:
    """
    This interface represents a link to an storage item.
    """

    property_deleted = "deleted"
    property_metadata = "metadata"
    property_object = "object"
    property_mutation = "mutation"
    id_special_prefix = "--gwbackupy-"

    # def __init__(self):

    def id(self) -> str:
        pass

    def is_special_id(self) -> bool:
        return self.id().startswith(LinkInterface.id_special_prefix)

    def get_properties(self) -> dict[str, any]:
        pass

    def set_properties(
        self, sets: dict[str, any], replace: bool = False
    ) -> LinkInterface:
        pass

    def get_property(self, name: str, default: any = None) -> any:
        pass

    def has_property(self, name: str) -> bool:
        pass

    def mutation(self) -> str:
        pass

    def is_deleted(self) -> bool:
        pass

    def is_metadata(self) -> bool:
        pass

    def is_object(self) -> bool:
        pass


LinkFilter = Callable[[LinkInterface], bool]
LinkGroupBy = Callable[[LinkInterface], list]


class LinkList(list):
    """
    List of links. The list is allows to group and filter.
    """

    def __init__(self, iterable):
        super().__init__(item for item in iterable)

    def __setitem__(self, index, item):
        super().__setitem__(index, str(item))

    def insert(self, index, item):
        super().insert(index, item)

    def append(self, item):
        super().append(item)

    def extend(self, other):
        if isinstance(other, type(self)):
            super().extend(other)
        else:
            super().extend(item for item in other)

    def find(
        self, f: LinkFilter | None, g: LinkGroupBy | None = None
    ) -> LinkInterface | dict[str, any] | None:
        """Find in links. If there are multiple results for a key, the most recent mutation is selected."""
        result = {}
        use_group_by = g is not None
        found = 0
        if g is None:
            g = lambda _: [""]
        for link in self:
            link: LinkInterface
            if f is not None and not f(link):
                continue
            found += 1
            gks = g(link)
            local_result = result
            for i in range(len(gks)):
                if i == len(gks) - 1:
                    break
                gk = gks[i]
                if gk not in local_result:
                    local_result[gk] = {}
                local_result = local_result[gk]
            gk = gks[-1]
            if gk not in local_result:
                local_result[gk] = link
                continue
            result_mutation = local_result[gk].mutation()
            if result_mutation is None:
                result_mutation = ""
            item_mutation = link.mutation()
            if item_mutation is None:
                item_mutation = ""

            if result_mutation < item_mutation:
                local_result[gk] = link
        if use_group_by:
            return result
        if found > 0:
            return result[""]
        return None


class StorageInterface:
    """
    Storage interface with base storage functionality
    """

    def new_link(
        self,
        object_id: str,
        extension: str,
        created_timestamp: int | float | None = None,
    ) -> LinkInterface:
        pass

    def get(self, link: LinkInterface) -> IO[bytes]:
        pass

    def put(self, link: LinkInterface, data: Data) -> bool:
        pass

    def remove(self, link: LinkInterface, as_new_mutation: bool = True) -> bool:
        pass

    def find(self, f: LinkFilter | None = None) -> LinkList[LinkInterface]:
        pass



from __future__ import annotations

import logging

from googleapiclient.errors import HttpError

from gwbackupy.helpers import is_rate_limit_exceeded, random_string
from gwbackupy.process_helpers import sleep_kc
from gwbackupy.providers.gmail_service_provider import GmailServiceProvider
from gwbackupy.providers.gmail_service_wrapper_interface import (
    GmailServiceWrapperInterface,
)


class GapiGmailServiceWrapper(GmailServiceWrapperInterface):
    def __init__(
        self,
        service_provider: GmailServiceProvider,
        try_count: int = 5,
        try_sleep: int = 10,
        dry_mode: bool = False,
    ):
        self.try_count = try_count
        self.try_sleep = try_sleep
        self.service_provider = service_provider
        self.dry_mode = dry_mode

    def get_service_provider(self) -> GmailServiceProvider:
        return self.service_provider

    def get_labels(self, email: str) -> list[dict[str, any]]:
        with self.service_provider.get_service(email) as service:
            response = service.users().labels().list(userId="me").execute()
            return response.get("labels", [])

    def get_messages(self, email: str, q: str) -> dict[str, dict[str, any]]:
        with self.service_provider.get_service(email) as service:
            messages = {}
            count = 0
            next_page_token = None
            page = 1
            while True:
                logging.debug(f"Loading page {page}. from server...")
                data = (
                    service.users()
                    .messages()
                    .list(userId="me", pageToken=next_page_token, maxResults=10000, q=q)
                    .execute()
                )
                next_page_token = data.get("nextPageToken", None)
                page_message_count = len(data.get("messages", []))
                logging.debug(
                    f"Page {page} successfully loaded (messages count: {page_message_count} / next page token: {next_page_token})"
                )
                # print(data)
                # exit(-1)
                count = count + page_message_count
                for message in data.get("messages", []):
                    messages[message.get("id")] = message
                page += 1
                if data.get("nextPageToken") is None:
                    break
            return messages

    def get_message(
        self, email: str, message_id: str, message_format: str = "minimal"
    ) -> dict[str, any] | None:
        for i in range(self.try_count):
            try:
                with self.service_provider.get_service(email) as service:
                    result = (
                        service.users()
                        .messages()
                        .get(userId="me", id=message_id, format=message_format)
                        .execute()
                    )
                    return result
            except HttpError as e:
                if e.status_code == 404:
                    # message not found
                    return None
                if i == self.try_count - 1:
                    # last try
                    logging.exception(f"{message_id} message download failed: {e}")
                    raise e
                if is_rate_limit_exceeded(e):
                    logging.warning(
                        f"{message_id} rate limit exceeded, sleeping for {self.try_sleep} seconds"
                    )
                    sleep_kc(self.try_sleep)
                else:
                    raise e
            except TimeoutError as e:
                if i == self.try_count - 1:
                    # last try
                    logging.exception(f"{message_id} message download failed: {e}")
                    raise e

    def create_label(
        self, email: str, name: str, get_if_already_exists: bool = False
    ) -> dict[str, any]:
        if self.dry_mode:
            return {
                "name": name,
                "id": f"DRYMODE{random_string()}",
                "type": "user",
            }
        for i in range(self.try_count):
            try:
                with self.service_provider.get_service(email) as service:
                    result = (
                        service.users()
                        .labels()
                        .create(userId="me", body={"name": name})
                        .execute()
                    )
                    return result
            except HttpError as e:
                if e.status_code == 409:
                    # already exists
                    logging.debug(f"label ({name}) already exists")
                elif i == self.try_count - 1:
                    # last try
                    logging.exception(f"Label ({name}) create fail: {e}")
                    raise e
                elif is_rate_limit_exceeded(e):
                    logging.warning(
                        f"Label ({name}) create rate limit exceeded, sleeping for {self.try_sleep} seconds"
                    )
                    sleep_kc(self.try_sleep)
                else:
                    logging.warning(f"Next attempt to create label ({name})")
            except TimeoutError as e:
                if i == self.try_count - 1:
                    # last try
                    logging.exception(f"Label ({name}) create fail: {e}")
                    raise e

            if not get_if_already_exists:
                return {
                    "name": name,
                    "type": "user",
                }
            labels = self.get_labels(email)
            for label in labels:
                if label.get("name") == name:
                    return label
            raise Exception(f"Label ({name}) is already exists but cannot found it")

    def insert_message(self, email: str, data: dict[str, any]) -> dict[str, any]:
        if self.dry_mode:
            return {"id": f"DRYMODE{random_string()}"}

        for i in range(self.try_count):
            try:
                with self.service_provider.get_service(email) as service:
                    result = (
                        service.users()
                        .messages()
                        .insert(
                            userId="me",
                            internalDateSource="dateHeader",
                            body=data,
                        )
                        .execute()
                    )
                    return result
            except HttpError as e:
                if i == self.try_count - 1:
                    # last try
                    logging.exception(f"Message insert fail: {e}")
                    raise e
                elif is_rate_limit_exceeded(e):
                    logging.warning(
                        f"Message insert rate limit exceeded, sleeping for {self.try_sleep} seconds"
                    )
                    sleep_kc(self.try_sleep)
                else:
                    logging.warning(f"Next attempt to insert message")



from __future__ import annotations

from gwbackupy.storage.storage_interface import (
    LinkList,
)


class ServiceItem:
    def __init__(self, provider: ServiceProviderInterface, email: str, service):
        self.__provider = provider
        self.__service = service
        self.__email = email

    def __enter__(self):
        if self.__service is None:
            raise ValueError(f"Service is released ({self.__email})")
        return self.__service

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self.__provider.release_service(self.__email, self.__service)
        finally:
            self.__service = None


class ServiceProviderInterface:
    def release_service(self, email: str, service):
        """Release the service"""
        pass

    def get_service(self, email: str) -> ServiceItem:
        """Get the service"""
        pass

    def storage_links(self, links: LinkList):
        """Pass storage links for tokens and more functionality"""
        pass



from datetime import datetime
from builtins import bool

import tzlocal


class FilterInterface:
    def __init__(self):
        pass

    def date_from(self, dt: datetime):
        pass

    def date_to(self, dt: datetime):
        pass

    def is_deleted(self):
        pass

    def match(self, d: any) -> bool:
        pass



from setuptools import setup
import gwbackupy.global_properties as global_properties

with open("README.md", "r") as fh:
    long_description = fh.read()

requires = []
with open("requirements.txt", "r") as fh:
    for line in fh.readlines():
        requires.append(line.strip())

setup(
    name="gwbackupy",
    version=global_properties.version,
    packages=[
        "gwbackupy",
        "gwbackupy.storage",
        "gwbackupy.filters",
        "gwbackupy.providers",
    ],
    url="https://github.com/smartondev/gwbackupy",
    license='BSD 3-Clause "New" or "Revised" License',
    author="Márton Somogyi",
    author_email="info@smarton.dev",
    description="Open source Google Workspace backup solution.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=requires,
    scripts=["scripts/gwbackupy"],
    python_requires=">=3.7",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
)



from datetime import datetime, timezone
from typing import Union

import tzlocal

from gwbackupy.filters.filter_interface import FilterInterface
from gwbackupy.storage.storage_interface import LinkInterface


class GmailFilter(FilterInterface):
    def __init__(self):
        super().__init__()
        self.__date_from: Union[datetime, None] = None
        self.__date_to: Union[datetime, None] = None
        self.__is_deleted: bool = False
        self.__is_missing: bool = False

    def date_to(self, dt: datetime):
        self.__date_to = dt.astimezone(timezone.utc)

    def date_from(self, dt: datetime):
        self.__date_from = dt.astimezone(timezone.utc)

    def is_deleted(self):
        self.__is_deleted = True

    def is_missing(self):
        self.__is_missing = True

    def match(self, d: any) -> bool:
        d: dict[str, any]
        link: LinkInterface = d["link"]
        if link.is_object():
            return True

        if self.__date_to is not None:
            ts1 = self.__date_to.timestamp()
            ts2 = int(link.mutation()) / 1000.0
            if ts2 >= ts1:
                return False
        if self.__date_from is not None:
            ts1 = self.__date_from.timestamp()
            ts2 = int(link.mutation()) / 1000.0
            if ts2 < ts1:
                return False
        if not self.__is_deleted and not self.__is_missing:
            # no missing or deleted filter
            return True
        if self.__is_deleted and link.is_deleted():
            # deleted
            return True
        if self.__is_missing:
            ids_from_server: dict[str, any] = d["server-data"]
            if link.id() not in ids_from_server:
                # missing
                return True
        return False



from __future__ import annotations

import hashlib
import logging
import tempfile
import threading

from google.auth.exceptions import RefreshError
from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import build
from oauth2client.service_account import ServiceAccountCredentials

from gwbackupy.providers.service_provider_interface import (
    ServiceProviderInterface,
    ServiceItem,
)
from gwbackupy.storage.storage_interface import (
    LinkInterface,
    StorageInterface,
    LinkList,
)


class GapiServiceProvider(ServiceProviderInterface):
    object_id_token = LinkInterface.id_special_prefix + "token--"
    """object ID for access token save and load"""

    def __init__(
        self,
        service_name: str,
        version: str,
        scopes: [str],
        credentials_file_path: str | None = None,
        service_account_file_path: str | None = None,
        service_account_email: str | None = None,
        storage: StorageInterface | None = None,
    ):
        self.service_name = service_name
        self.version = version
        self.scopes = scopes
        self.credentials_file_path = credentials_file_path
        self.service_account_file_path = service_account_file_path
        self.service_account_email = service_account_email
        self.storage = storage
        self.tlock = threading.RLock()
        self.services: dict[str, list[any]] = dict()
        self.credentials_token_links: dict[str, LinkInterface] = dict()

    def release_service(self, email: str, service):
        logging.debug(f"Release service ({email})")
        if service is None:
            return
        with self.tlock:
            if email not in self.services:
                return
            self.services[email].append(service)

    def get_service(self, email: str):
        service = None
        with self.tlock:
            if email not in self.services.keys():
                self.services[email] = []
            if len(self.services[email]) > 0:
                logging.debug(f"Reuse service ({email})")
                service = self.services[email].pop()
            if service is None:
                logging.debug("Create new service")
                if self.credentials_file_path is not None:
                    credentials = self.__get_credentials_by_oauth(email)
                elif self.service_account_file_path is not None:
                    credentials = self.__get_credentials_by_service_account(email)
                else:
                    raise Exception(f"Not supported credentials")
                if not credentials:
                    raise Exception(f"Credentials cannot be None")
                service = build(
                    self.service_name,
                    self.version,
                    credentials=credentials,
                )
        return ServiceItem(self, email, service)

    def storage_links(self, links: LinkList):
        """Filter stored tokens for OAuth authentication"""
        self.credentials_token_links = links.find(
            f=lambda l: l.id() == GapiServiceProvider.object_id_token,
            g=lambda l: [l.get_properties().get("email", "")],
        )

    def __get_credentials_by_service_account(self, email: str):
        """get credentials by service account access"""
        extension = self.service_account_file_path.split(".")[-1].lower()
        if extension == "p12":
            if (
                self.service_account_email is None
                or self.service_account_email.strip() == ""
            ):
                raise Exception("Service account email is required for p12 keyfile")
            credentials = ServiceAccountCredentials.from_p12_keyfile(
                self.service_account_email,
                self.service_account_file_path,
                "notasecret",
                scopes=self.scopes,
            )
        elif extension == "json":
            credentials = ServiceAccountCredentials.from_json_keyfile_name(
                self.service_account_file_path,
                self.scopes,
            )
            pass
        else:
            raise Exception(f"Not supported service account file extension")

        credentials = credentials.create_delegated(email)
        return credentials

    def __get_credentials_by_oauth(self, email: str):
        """Get credentials object from OAuth access. This method store token and reuse/refresh if required"""
        credentials = None
        fd, temp = tempfile.mkstemp()
        email_md5 = hashlib.md5(email.encode("utf-8")).hexdigest().lower()
        if email_md5 in self.credentials_token_links:
            logging.debug("Try to load previously saved token")
            token_link = self.credentials_token_links.get(email_md5)
            with self.storage.get(token_link) as tf:
                with open(temp, "wb") as tfo:
                    tfo.write(tf.read())
            credentials = Credentials.from_authorized_user_file(temp)
        if not credentials or not credentials.valid:
            logging.debug("Credentials not found or not valid")
            if credentials and credentials.expired and credentials.refresh_token:
                logging.debug("Try to refresh token")
                try:
                    credentials.refresh(Request())
                except RefreshError as e:
                    logging.exception(f"Failed to refresh token: {e}")
                    credentials = None
            if not credentials:
                logging.debug("Credentials not found")
                flow = InstalledAppFlow.from_client_secrets_file(
                    self.credentials_file_path,
                    self.scopes,
                )
                credentials = flow.run_local_server(port=0)

            token_link_new = self.storage.new_link(
                GapiServiceProvider.object_id_token, "json"
            )
            token_link_new.set_properties({"email": email_md5})
            logging.debug(f"Put token to storage ({token_link_new})")
            result = self.storage.put(token_link_new, credentials.to_json())
            if not result:
                raise Exception(f"{email} Failed to store token ({token_link_new})")
            else:
                logging.info("token stored successfully")
                token_link_old = self.credentials_token_links.get(email_md5)
                self.credentials_token_links[email_md5] = token_link_new
                if token_link_old:
                    result = self.storage.remove(token_link_old, False)
                    if result:
                        logging.debug(f"{email} Old token removed successfully")
                    else:
                        logging.warning(f"{email} Old token removed fail")
        return credentials



from __future__ import annotations

from gwbackupy.providers.service_provider_interface import ServiceProviderInterface


class GmailServiceWrapperInterface:
    def get_messages(self, email: str, q: str) -> dict[str, [dict[str, any]]]:
        """Return all messages that match with query"""
        pass

    def get_message(
        self, email: str, message_id: str, message_format: str = "minimal"
    ) -> dict[str, any] | None:
        """Get one message by message ID. If not exists then return None"""
        pass

    def get_labels(self, email: str) -> list[dict[str, any]]:
        """Return all labels"""
        pass

    def get_service_provider(self) -> ServiceProviderInterface:
        pass

    def create_label(
        self, email: str, name: str, get_if_already_exists: bool = False
    ) -> dict[str, any]:
        """Create a label if not existing, and return the label data"""
        pass

    def insert_message(self, email: str, data: dict[str, any]) -> dict[str, any]:
        """Insert a message to server with specified data, and return new message ID"""
        pass



from __future__ import annotations

import collections
import concurrent.futures
import gzip
import json
import logging
import threading
from datetime import datetime, timedelta

from gwbackupy import global_properties
from gwbackupy.filters.filter_interface import FilterInterface
from gwbackupy.helpers import (
    decode_base64url,
    encode_base64url,
    str_trim,
    json_load,
)
from gwbackupy.process_helpers import is_killed, sleep_kc, await_all_futures
from gwbackupy.providers.gmail_service_wrapper_interface import (
    GmailServiceWrapperInterface,
)
from gwbackupy.storage.storage_interface import (
    StorageInterface,
    LinkList,
    LinkInterface,
    Data,
)


class Gmail:
    """Gmail service"""

    object_id_labels = "--gwbackupy-labels--"
    """Gmail's special object ID for storing labels"""

    def __init__(
        self,
        email: str,
        storage: StorageInterface,
        service_wrapper: GmailServiceWrapperInterface,
        batch_size: int = 10,
        labels: list[str] | None = None,
        dry_mode: bool = False,
    ):
        self.dry_mode = dry_mode
        self.email = email
        self.storage = storage
        if batch_size is None or batch_size < 1:
            batch_size = 5
        self.batch_size = batch_size
        self.__lock = threading.RLock()
        self.__services = {}
        self.__error_count = 0
        self.__service_wrapper = service_wrapper
        if labels is None:
            labels = []
        self.labels = labels

    def __get_local_messages_latest_mutations_only(self):
        pass

    def __get_message_from_server(self, message_id, message_format="raw", email=None):
        if email is None:
            email = self.email
        logging.debug(
            f"{message_id} download from server with format: {message_format}"
        )
        result = self.__service_wrapper.get_message(email, message_id, message_format)
        logging.debug(f"{message_id} successfully downloaded")
        return result

    def __store_message_file(
        self, message_id: str, raw_message: bytes, create_timestamp: float
    ):
        logging.debug("Store message {id}".format(id=message_id))
        link = self.storage.new_link(
            object_id=message_id, extension="eml.gz", created_timestamp=create_timestamp
        ).set_properties({LinkInterface.property_object: True})
        result = self.__storage_put(
            link, data=gzip.compress(raw_message, compresslevel=9)
        )
        if result:
            logging.info(f"{message_id} message is saved")
        else:
            raise Exception("Mail message save failed")

    def __backup_messages(
        self, message, stored_messages: dict[str, dict[int, LinkInterface]]
    ):
        message_id = message.get("id", "UNKNOWN")  # for logging
        try:
            # if message_id != '1853ee437c8ff302':
            #     raise Exception('SKIP')
            message_id = message["id"]
            latest_meta_link = None
            if message_id in stored_messages:
                latest_meta_link = stored_messages[message_id][0]
            is_new = latest_meta_link is None
            if is_new:
                logging.debug(f"{message_id} is new")
            # TODO: option for force raw mode
            message_format = "raw"
            if not is_new and stored_messages[message_id][1] is not None:
                message_format = "minimal"
            data = self.__get_message_from_server(message_id, message_format)
            if data is None:
                # (deleted)
                logging.info(f"{message_id} is not found on server")
                return

            subject = str_trim(data.get("snippet", ""), 64)
            if is_new:
                logging.info(f"{message_id} new message, snippet: {subject}")
            else:
                logging.debug(f"{message_id} Snippet: {subject}")

            create_timestamp = int(data["internalDate"]) / 1000.0
            if "raw" in data.keys():
                raw = decode_base64url(data.get("raw"))
                self.__store_message_file(message_id, raw, create_timestamp)
                data.pop("raw")

            write_meta = True  # if any failure then write it force
            if not is_new:
                logging.log(
                    global_properties.log_finest,
                    f"{message_id} load local version of meta data",
                )
                try:
                    with self.storage.get(latest_meta_link) as mf:
                        d = json.load(mf)
                        logging.log(
                            global_properties.log_finest,
                            f"{message_id} metadata is loaded from local",
                        )
                        if d == data:
                            write_meta = False
                except BaseException as e:
                    logging.exception(f"{message_id} metadata load as json failed: {e}")

            if write_meta:
                link = self.storage.new_link(
                    object_id=message_id,
                    extension="json",
                    created_timestamp=create_timestamp,
                ).set_properties({LinkInterface.property_metadata: True})
                success = self.__storage_put(link, data=json.dumps(data))
                if success:
                    logging.info(f"{message_id} meta data is saved")
                else:
                    raise Exception("Meta data put failed")
            else:
                logging.debug(f"{message_id} meta data is not changed, skip put")
            if message_id in stored_messages:
                del stored_messages[message_id]
        except Exception as e:
            with self.__lock:
                self.__error_count += 1
            if str(e) == "SKIP":
                return
            logging.exception(f"{message_id} {e}")

    def __get_all_messages_from_server(
        self, email: str | None = None, q: str | None = "label:all"
    ):
        if email is None:
            email = self.email
        logging.info("Get all message ids from server...")
        messages = self.__service_wrapper.get_messages(email, q)
        logging.info(f"Message(s) count: {len(messages)}")
        # print(count)
        # print(messages)
        return messages

    def __get_labels_from_server(self, email=None) -> list[dict[str, any]]:
        if email is None:
            email = self.email
        logging.info(f"Getting labels from server ({email})")
        return self.__service_wrapper.get_labels(email)

    def __backup_labels(self, link: LinkInterface | None) -> bool:
        logging.info("Backing up labels...")
        labels = self.__get_labels_from_server()
        if link is not None:
            logging.debug("labels is exists, checking for changes")
            try:
                with self.storage.get(link) as f:
                    d = json.load(f)
                    if d == labels:
                        logging.info("Labels is not changed, not saving it")
                        return True
                    else:
                        logging.debug("labels is changed")
            except BaseException as e:
                logging.exception(f"labels loading or parsing failed: {e}")
        link = self.storage.new_link(
            object_id=Gmail.object_id_labels, extension="json", created_timestamp=None
        ).set_properties({LinkInterface.property_metadata: True})
        result = self.__storage_put(link, data=json.dumps(labels))
        if not result:
            logging.error("Error while storing labels")
            return False
        logging.info("Backing up labels successfully")
        return True

    def __storage_remove(self, link: LinkInterface) -> bool:
        if self.dry_mode:
            logging.info(f"DRY MODE storage remove: {link}")
            return True
        return self.storage.remove(link)

    def __storage_put(self, link: LinkInterface, data: Data) -> bool:
        if self.dry_mode:
            logging.info(f"DRY MODE storage put: {link}")
            return True
        return self.storage.put(link, data)

    def backup(self, quick_sync_days: int | None = None) -> bool:
        logging.info(f"Starting backup for {self.email}")
        self.__error_count = 0

        logging.info("Scanning backup storage...")
        stored_data_all = self.storage.find()
        logging.info(f"Stored items: {len(stored_data_all)}")
        self.__service_wrapper.get_service_provider().storage_links(stored_data_all)

        labels_link = stored_data_all.find(
            f=lambda l: l.id() == Gmail.object_id_labels and l.is_metadata
        )
        if not self.__backup_labels(labels_link):
            logging.error("Backup finished with storing labels failed")
            return False
        if quick_sync_days is not None and quick_sync_days < 1:
            quick_sync_days = None
        if quick_sync_days is not None:
            logging.info(f"Quick syncing, going back {quick_sync_days} days")

        stored_messages: dict[str, dict[int, LinkInterface]] = stored_data_all.find(
            f=lambda l: not l.is_special_id() and (l.is_metadata() or l.is_object()),
            g=lambda l: [l.id(), 0 if l.is_metadata() else 1],
        )
        del stored_data_all
        for message_id in list(stored_messages.keys()):
            link_metadata = stored_messages[message_id].get(0)
            if link_metadata is None:
                logging.error(f"{message_id} metadata is not found in locally")
                del stored_messages[message_id]
            elif link_metadata.is_deleted():
                logging.debug(f"{message_id} metadata is already deleted")
                del stored_messages[message_id]
            else:
                logging.log(
                    global_properties.log_finest,
                    f"{message_id} is usable from backup storage",
                )
        logging.info(f"Stored messages: {len(stored_messages)}")

        q = "label:all"
        if quick_sync_days is not None:
            date = datetime.now() - timedelta(days=quick_sync_days)
            q = f"label:all after:{date.strftime('%Y/%m/%d')}"
        messages_from_server = self.__get_all_messages_from_server(q=q)
        logging.info("Processing...")
        executor = concurrent.futures.ThreadPoolExecutor(max_workers=self.batch_size)
        futures = []
        # submit message download jobs
        for message_id in messages_from_server:
            futures.append(
                executor.submit(
                    self.__backup_messages,
                    messages_from_server[message_id],
                    stored_messages,
                )
            )
        # wait for jobs
        if not await_all_futures(futures):
            # cancel jobs
            executor.shutdown(cancel_futures=True)
            logging.warning("Process is killed")
            return False
        logging.info("Processed")

        if self.__error_count > 0:
            # if error then never delete!
            logging.error("Backup failed with " + str(self.__error_count) + " errors")
            return False

        if quick_sync_days is None:
            logging.info("Mark as deletes...")
            for message_id in stored_messages:
                if is_killed():
                    logging.warning("Process is killed")
                    return False

                links = stored_messages[message_id]
                logging.debug(f"{message_id} mark as deleted in local storage...")
                meta_link = links.get(0)
                if meta_link is None:
                    continue
                logging.debug(f"{message_id} - {meta_link}")
                if self.__storage_remove(meta_link):
                    logging.debug(f"{message_id} metadata mark as deleted successfully")
                    message_link = links.get(1)
                    if message_link is None:
                        logging.info(f"{message_id} marked as deleted")
                    else:
                        if self.__storage_remove(message_link):
                            logging.debug(
                                f"{message_id} object mark as deleted successfully"
                            )
                            logging.info(f"{message_id} marked as deleted")
                        else:
                            logging.error(f"{message_id} object mark as deleted fail")
                else:
                    logging.error(f"{message_id} mark as deleted failed")
            logging.info("Mark as deleted: complete")
        else:
            logging.info("Quick syncing mode, skip deletion for locale storage")
        logging.info(f"Backup finished for {self.email}")
        return True

    def __create_label_server(self, label_name, email) -> dict:
        if email is None:
            email = self.email
        logging.info(f"Restoring label if not exists: {label_name}")
        if self.dry_mode:
            logging.info("DRY MODE: create label if not exists")
            return {
                "id": f"Label_DRY{datetime.utcnow().timestamp():1.3f}",
                "type": "user",
                "name": label_name,
            }
        return self.__service_wrapper.create_label(email=email, name=label_name)

    def __get_restore_label_ids(
        self,
        to_email: str,
        labels_from_storage: dict[str, dict[str, any]],
        labels_form_server: dict[str, dict[str, any]],
        add_labels: [str],
        label_ids_from_message: [str],
    ) -> [str]:
        if self.email != to_email:
            raise NotImplementedError("Not implemented for different emails")
        with self.__lock:
            result = []
            for message_label_id in label_ids_from_message:
                if message_label_id == "CHAT":
                    # CHAT tag cannot be restoring
                    continue
                if message_label_id in labels_form_server:
                    server_data = labels_form_server[message_label_id]
                    if server_data["type"] == "system" or server_data["type"] == "user":
                        # user and system tag allow directly if exists on server
                        result.append(server_data["id"])
                        continue
                    raise NotImplementedError(
                        f'Not implemented tag type: {server_data["type"]}'
                    )
                # not exists on server
                if message_label_id not in labels_from_storage:
                    logging.warning(
                        f"Label with {message_label_id} ID is cannot be restored because no data can be found."
                    )
                    continue
                label_data = labels_from_storage[message_label_id]
                created_label_data = self.__create_label_server(
                    label_data["name"], to_email
                )
                if created_label_data is None:
                    raise Exception(f"Label is created already? A ({label_data})")
                # label data stored on original label ID!
                labels_form_server[message_label_id] = created_label_data
                result.append(created_label_data["id"])
            for add_label in add_labels:
                found = False
                for key in labels_form_server:
                    server_data = labels_form_server[key]
                    if server_data["name"] == add_label:
                        result.append(server_data["id"])
                        found = True
                        break
                if found:
                    continue
                # create label...
                created_label_data = self.__create_label_server(add_label, to_email)
                if created_label_data is None:
                    raise Exception(f"Label is created already? B ({label_data})")
                # label data stored on original label ID!
                labels_form_server[created_label_data["id"]] = created_label_data
                result.append(created_label_data["id"])

            return result

    def __restore_message(
        self,
        restore_message_id: str,
        link: dict[int, LinkInterface],
        to_email: str,
        labels_from_storage: dict[str, dict[str, any]],
        labels_form_server: dict[str, dict[str, any]],
        add_labels: [str],
    ):
        try:
            logging.info(f"{restore_message_id} Restoring message...")
            if 0 not in link or 1 not in link:
                raise Exception("Metadata and/or message link not found in storage")
            with self.storage.get(link[0]) as mf:
                meta = json.load(mf)
            logging.debug(f"{restore_message_id} {meta}")
            with self.storage.get(link[1]) as mf:
                message_content = gzip.decompress(mf.read())
            label_ids_from_message: [str] = meta["labelIds"]
            try:
                label_ids_from_message.index("CHAT")
                # not restorable as message
                logging.info(
                    f"{restore_message_id} message with CHAT label is not supported."
                )
                return
            except ValueError:
                pass
            label_ids = self.__get_restore_label_ids(
                to_email=to_email,
                labels_from_storage=labels_from_storage,
                labels_form_server=labels_form_server,
                add_labels=add_labels,
                label_ids_from_message=label_ids_from_message,
            )
            label_names = []
            for label_id in meta["labelIds"]:
                label_names.append(labels_form_server[label_id]["name"])

            logging.info(
                f"{restore_message_id} snippet: {str_trim(meta['snippet'], 80)} / labels: {', '.join(label_names)}"
            )

            message_data = {
                "labelIds": label_ids,
                "raw": encode_base64url(message_content),
            }
            subject = str_trim(meta.get("snippet", ""), 64)
            if self.dry_mode:
                logging.info(f"DRY MODE {restore_message_id} message insert")
            result = self.__service_wrapper.insert_message(to_email, message_data)
            logging.debug(f"Message uploaded {result}")
            logging.info(
                f'{restore_message_id}->{result.get("id")} Message uploaded ({subject})'
            )
            return result
        except BaseException as e:
            with self.__lock:
                self.__error_count += 1
            logging.exception(f"{restore_message_id} Exception: {e}")

    def __load_labels_from_storage(
        self, links: LinkList[LinkInterface]
    ) -> dict[str, dict[str, any]] | None:
        logging.info(f"Loading labels...")
        labels_links: dict[str, LinkInterface] = links.find(
            f=lambda l: l.id() == Gmail.object_id_labels and l.is_metadata,
            g=lambda l: [l.mutation()],
        )
        labels_links = collections.OrderedDict(sorted(labels_links.items()))
        result = {}
        for key in labels_links:
            labels_link = labels_links[key]
            with self.storage.get(labels_link) as lf:
                json_data: list[dict[str, any]] = json_load(lf)
                if json_data is None:
                    logging.error(f"Stored labels read from JSON failed")
                    return None
                for item in json_data:
                    if item.get("id") is None:
                        logging.error(f"Invalid structure of stored labels")
                        return None
                    result[item.get("id")] = item
        logging.info(f"Labels loaded successfully ({len(result)})")
        return result

    def restore(
        self,
        item_filter: FilterInterface,
        to_email: str | None = None,
        restore_deleted: bool = False,
        add_labels: list[str] | None = None,
        restore_missing: bool = False,
    ):
        self.__error_count = 0
        if to_email is None:
            to_email = self.email

        if not restore_deleted and not restore_missing:
            logging.warning("Tasks not found, see more e.g. --restore-deleted")
            return True

        logging.info("Scanning backup storage...")
        stored_data_all = self.storage.find()
        logging.info(f"Stored items: {len(stored_data_all)}")
        self.__service_wrapper.get_service_provider().storage_links(stored_data_all)

        latest_labels_from_storage = self.__load_labels_from_storage(stored_data_all)
        if latest_labels_from_storage is None:
            logging.error("Stored labels loading failed")
            return False
        _labels_from_server = self.__get_labels_from_server()
        if _labels_from_server is None:
            logging.error("Loading labels from server failed")
            return False
        labels_from_server = {}
        for label in _labels_from_server:
            if "id" not in label:
                raise Exception("Not supported label structure")
            labels_from_server[label.get("id")] = label
        del _labels_from_server
        messages_from_server_dest_email = {}
        if self.email == to_email:
            messages_from_server_dest_email = self.__get_all_messages_from_server()

        logging.info("Filtering messages...")
        stored_messages: dict[str, dict[int, LinkInterface]] = stored_data_all.find(
            f=lambda l: not l.is_special_id()
            and (l.is_metadata() or l.is_object())
            and item_filter.match(
                {
                    "message-id": l.id(),
                    "link": l,
                    "server-data": messages_from_server_dest_email,
                }
            ),
            g=lambda l: [l.id(), 0 if l.is_metadata() else 1],
        )
        del stored_data_all
        for message_id in list(stored_messages.keys()):
            if (
                stored_messages[message_id].get(0) is None
                or stored_messages[message_id].get(1) is None
            ):
                # no metadata or no object
                del stored_messages[message_id]

        logging.info(f"Number of potentially affected messages: {len(stored_messages)}")
        logging.info("Upload messages...")
        executor = concurrent.futures.ThreadPoolExecutor(max_workers=self.batch_size)
        futures = []
        for message_id in stored_messages:
            futures.append(
                executor.submit(
                    self.__restore_message,
                    message_id,
                    stored_messages[message_id],
                    to_email,
                    latest_labels_from_storage,
                    labels_from_server,
                    add_labels,
                )
            )
        if not await_all_futures(futures):
            logging.warning("Process killed")
            return False

        if self.__error_count > 0:
            logging.error(f"Messages uploaded with {self.__error_count} errors")
            return False
        logging.info("Messages uploaded successfully")

        return True
